@@ -39,10 +39,18 @@
 #include <libedataserver/e-source.h>
 #include <libedataserver/e-source-list.h>
 #include <libedataserver/e-source-group.h>
+#include <libedataserver/e-xml-hash-utils.h>
 
+#include <camel/camel-url.h>
+
+#include <sys/types.h>
+#include <dirent.h>
+#include <sys/stat.h>
+#include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 
+
 struct _ExchangeConfigListenerPrivate {
 	GConfClient *gconf;
 	guint idle_id;
@@ -84,6 +92,10 @@ static void account_changed (EAccountList *account_listener,
 static void account_removed (EAccountList *account_listener,
 			     EAccount     *account);
 
+static gboolean exchange_camel_urls_is_equal (const gchar *url1, 
+					      const gchar *url2);
+
+
 static void
 class_init (GObjectClass *object_class)
 {
@@ -168,52 +180,6 @@ E2K_MAKE_TYPE (exchange_config_listener, ExchangeConfigListener, class_init, ini
 #define EVOLUTION_URI_PREFIX     "evolution:/"
 #define EVOLUTION_URI_PREFIX_LEN (sizeof (EVOLUTION_URI_PREFIX) - 1)
 
-static EFolder *
-standard_folder (ExchangeAccount *account, const char *folder_type)
-{
-	const char *uri;
-
-	uri = exchange_account_get_standard_uri (account, folder_type);
-	if (!uri)
-		return NULL;
-	return exchange_account_get_folder (account, uri);
-}
-
-static void
-set_special_mail_folder (ExchangeAccount *account, const char *folder_type,
-			 char **physical_uri)
-{
-	EFolder *folder;
-
-	folder = standard_folder (account, folder_type);
-	if (!folder)
-		return;
-
-	g_free (*physical_uri);
-	*physical_uri = g_strdup (e_folder_get_physical_uri (folder));
-}
-
-static void
-add_defaults_for_account (ExchangeConfigListener *config_listener,
-			  E2kContext *ctx,
-			  ExchangeAccount *account)
-{
-	EAccount *eaccount;
-
-#if LDEAD
-	add_autocompletion_folders (config_listener->priv->gconf, account);
-#endif
-
-	eaccount = config_listener->priv->configured_account;
-	set_special_mail_folder (account, "drafts",
-				 &eaccount->drafts_folder_uri);
-	set_special_mail_folder (account, "sentitems",
-				 &eaccount->sent_folder_uri);
-	e_account_list_change (E_ACCOUNT_LIST (config_listener), eaccount);
-	e_account_list_save (E_ACCOUNT_LIST (config_listener));
-}
-
-
 static gboolean
 is_active_exchange_account (EAccount *account)
 {
@@ -225,20 +191,122 @@ is_active_exchange_account (EAccount *account)
 }
 
 static void 
-remove_account_esource (ExchangeAccount *account, 
+update_foreign_uri (const char *path, const char *account_uri)
+{
+	char *file_path, *phy_uri, *foreign_uri, *new_phy_uri;
+	struct stat file_stat;
+	GHashTable *old_props = NULL;
+	xmlDoc *old_doc, *new_doc = NULL;
+
+	if (!path)
+		return;
+
+	file_path = g_build_filename (path, "hierarchy.xml", NULL);
+	if (stat (file_path, &file_stat) < 0)
+		goto cleanup;
+
+	old_doc = xmlParseFile (file_path);
+	if (!old_doc)
+		goto cleanup;
+
+	old_props = e_xml_to_hash (old_doc, E_XML_HASH_TYPE_PROPERTY);
+	xmlFreeDoc (old_doc);
+
+	phy_uri = g_hash_table_lookup (old_props, "physical_uri_prefix");
+	if (!phy_uri)
+		goto cleanup;
+
+	foreign_uri = strstr (phy_uri, "://");
+	if (!foreign_uri)
+		goto cleanup;
+	foreign_uri = strchr (foreign_uri + 3, '/');
+	if (!foreign_uri)
+		goto cleanup;
+
+	if ((foreign_uri + 1) && (*(foreign_uri + 1) == ';'))
+		goto cleanup;
+
+	new_phy_uri = g_strdup_printf ("exchange://%s/;%s", account_uri, foreign_uri + 1);
+	g_hash_table_remove (old_props, "physical_uri_prefix");
+	g_hash_table_insert (old_props, (char *)g_strdup ("physical_uri_prefix"), new_phy_uri);
+
+	new_doc = e_xml_from_hash (old_props, E_XML_HASH_TYPE_PROPERTY, "foreign-hierarchy");
+	xmlSaveFile (file_path, new_doc);
+
+	xmlFreeDoc (new_doc);
+	g_free (new_phy_uri);
+cleanup:
+	g_free (file_path);
+	g_hash_table_destroy (old_props);
+	return;
+}
+
+static void
+migrate_foreign_hierarchy (ExchangeAccount *account)
+{
+	DIR *d;
+	struct dirent *dentry;
+	char *dir;
+
+	d = opendir (account->storage_dir);
+	if (d) {
+		while ((dentry = readdir (d))) {
+			if (!strchr (dentry->d_name, '@'))
+				continue;
+			dir = g_strdup_printf ("%s/%s", account->storage_dir, 
+							dentry->d_name);
+			update_foreign_uri (dir, account->account_filename);
+			g_free (dir);
+		}
+		closedir (d);
+	}
+}
+
+static void
+ex_set_relative_uri (ESource *source, const char *url)
+{
+	const char *rel_uri = e_source_peek_relative_uri (source);
+	char *folder_name;
+	char *new_rel_uri;
+
+	if (!rel_uri)
+		return;
+
+	folder_name = strchr (rel_uri, '/');
+	if (!folder_name)
+		return;
+
+	if ((folder_name + 1) && *(folder_name + 1) == ';') 
+		return;
+
+	new_rel_uri = g_strdup_printf ("%s;%s", url, folder_name + 1);
+	e_source_set_relative_uri (source, new_rel_uri);
+}
+
+static void 
+migrate_account_esource (EAccount *account, 
 		        FolderType folder_type)
 {
 	ESourceGroup *group;
 	ESource *source = NULL;
 	GSList *groups;
 	GSList *sources;
-	GSList *ids, *node_to_be_deleted;
 	gboolean found_group;
-	const char *source_uid;
+	const char *user_name, *authtype;
 	GConfClient *client;
 	ESourceList *source_list = NULL;
+	CamelURL *camel_url;
+	char *url_string;
 
-	/* Remove the ESource group, to remove all the folders in a component */
+	camel_url = camel_url_new (account->source->url, NULL);
+	if (!camel_url)
+		return;
+	user_name = camel_url->user;
+	authtype  = camel_url->authmech;
+	url_string = camel_url_to_string (camel_url, CAMEL_URL_HIDE_PASSWORD | CAMEL_URL_HIDE_PARAMS);
+
+	if (!user_name) 
+		return;
 
 	client = gconf_client_get_default ();
 
@@ -258,63 +326,25 @@ remove_account_esource (ExchangeAccount *account,
 	for ( ; groups != NULL && !found_group; groups = g_slist_next (groups)) {
 		group = E_SOURCE_GROUP (groups->data);
 
-		if (strcmp (e_source_group_peek_name (group), account->account_name) == 0
+		if (strcmp (e_source_group_peek_name (group), account->name) == 0
                     &&
                     strcmp (e_source_group_peek_base_uri (group), EXCHANGE_URI_PREFIX) == 0) {
 			sources = e_source_group_peek_sources (group);
 
+			found_group = TRUE;
 			for( ; sources != NULL; sources = g_slist_next (sources)) {
 				source = E_SOURCE (sources->data);
-				source_uid = e_source_peek_uid (source);
 
-				/* Remove from the selected folders */
-				if (folder_type == EXCHANGE_CALENDAR_FOLDER) {
-					ids = gconf_client_get_list (
-							client, 
-							CONF_KEY_SELECTED_CAL_SOURCES , 
-							GCONF_VALUE_STRING, NULL);
-					if (ids) {
-						node_to_be_deleted = g_slist_find_custom (
-									ids, 
-									source_uid, 
-									(GCompareFunc) strcmp);
-						if (node_to_be_deleted) {
-							g_free (node_to_be_deleted->data);
-							ids = g_slist_delete_link (ids, 
-									node_to_be_deleted);
-							gconf_client_set_list (client, 
-								CONF_KEY_SELECTED_CAL_SOURCES,
-								GCONF_VALUE_STRING, ids, NULL);
-						}
-						g_slist_foreach (ids, (GFunc) g_free, NULL);
-						g_slist_free (ids);
-					}
-				}
-				else if (folder_type == EXCHANGE_TASKS_FOLDER) {
-					ids = gconf_client_get_list (client, 
-								CONF_KEY_SELECTED_TASKS_SOURCES , 
-								GCONF_VALUE_STRING, NULL);
-					if (ids) {
-						node_to_be_deleted = g_slist_find_custom (
-									ids,
-									source_uid, 
-									(GCompareFunc) strcmp);
-						if (node_to_be_deleted) {
-							g_free (node_to_be_deleted->data);
-							ids = g_slist_delete_link (ids, 
-									node_to_be_deleted);
-							gconf_client_set_list (client, 
-								CONF_KEY_SELECTED_TASKS_SOURCES,
-								GCONF_VALUE_STRING, ids, NULL);
-						}
-						g_slist_foreach (ids, (GFunc) g_free, NULL);
-						g_slist_free (ids);
-					}
-				}
-				e_source_list_remove_group (source_list, group);
+				ex_set_relative_uri (source, url_string + strlen ("exchange://"));
+				e_source_set_property (source, "username", user_name);
+				e_source_set_property (source, "auth-domain", "Exchange");
+				if (authtype)
+					e_source_set_property (source, "auth-type", authtype);
+				if (folder_type == EXCHANGE_CONTACTS_FOLDER)
+					e_source_set_property (source, "auth", "plain/password");
+				else
+					e_source_set_property (source, "auth", "1");
 				e_source_list_sync (source_list, NULL);
-				found_group = TRUE;
-				break;
 			}
 		}
 	}
@@ -322,14 +352,17 @@ remove_account_esource (ExchangeAccount *account,
 	g_object_unref (client);
 }
 
-static void
-remove_account_esources (ExchangeAccount *account)
+void
+exchange_config_listener_migrate_esources (ExchangeConfigListener *config_listener)
 {
-	/* Remove ESources for all the folders in all the components */
+	g_return_if_fail (config_listener != NULL);
 
-	remove_account_esource (account, EXCHANGE_CALENDAR_FOLDER);
-	remove_account_esource (account, EXCHANGE_TASKS_FOLDER);
-	remove_account_esource (account, EXCHANGE_CONTACTS_FOLDER);
+	EAccount *account = config_listener->priv->configured_account;
+
+	migrate_account_esource (account, EXCHANGE_CALENDAR_FOLDER);
+	migrate_account_esource (account, EXCHANGE_TASKS_FOLDER);
+	migrate_account_esource (account, EXCHANGE_CONTACTS_FOLDER);
+	migrate_foreign_hierarchy (config_listener->priv->exchange_account);
 }
 
 static void
@@ -337,7 +370,6 @@ account_added (EAccountList *account_list, EAccount *account)
 {
 	ExchangeConfigListener *config_listener;
 	ExchangeAccount *exchange_account;
-	ExchangeAccountResult result;
 
 	if (!is_active_exchange_account (account))
 		return;
@@ -345,8 +377,6 @@ account_added (EAccountList *account_list, EAccount *account)
 	config_listener = EXCHANGE_CONFIG_LISTENER (account_list);
 	if (config_listener->priv->configured_account) {
 		/* Multiple accounts configured. */
-		e_notice (NULL, GTK_MESSAGE_ERROR,
-			  _("You may only configure a single Exchange account"));
 		return;
 	}
 
@@ -366,16 +396,10 @@ account_added (EAccountList *account_list, EAccount *account)
 	g_free (config_listener->priv->configured_name);
 	config_listener->priv->configured_name = g_strdup (account->name);
 
-	if (account == e_account_list_get_default (account_list)) {
-		g_signal_connect_swapped (config_listener->priv->exchange_account,
-					  "connected",
-					  G_CALLBACK (add_defaults_for_account),
-					  config_listener);
-	}
 
 	g_signal_emit (config_listener, signals[EXCHANGE_ACCOUNT_CREATED], 0,
 		       exchange_account);
-	exchange_account_connect (exchange_account, NULL, &result);
+	exchange_config_listener_migrate_esources (config_listener);
 }
 
 struct account_update_data {
@@ -479,7 +503,12 @@ account_changed (EAccountList *account_list, EAccount *account)
 		return;
 	}
 
-	if (!strcmp (config_listener->priv->configured_uri, account->source->url) &&
+	/* FIXME: The order of the parameters in the Camel URL string is not in 
+	 * order for the two given strings. So, we will not be able to use
+	 * plain string comparison. Instead compare the parameters one by one.
+	 */
+	if (exchange_camel_urls_is_equal (config_listener->priv->configured_uri, 
+					  account->source->url) &&
 	    !strcmp (config_listener->priv->configured_name, account->name)) {
 		/* The user changed something we don't care about. */
 		return;
@@ -513,42 +542,23 @@ account_changed (EAccountList *account_list, EAccount *account)
 	 * could be due to change in hostname or some parameter value, 
 	 * remove old e-sources 
 	 */
-	if (strcmp (config_listener->priv->configured_uri, account->source->url)) {
-		/* This condition will be true always as order of parameters in 
-		 * configured_uri and source->url will not match
-		 */
-		remove_account_esources (priv->exchange_account);
-
-		/* Ask user to authenticate at next login if username, hostname,
-		 * OWA URL or GC server values are changed. 
-		 */
-		if (requires_relogin (config_listener->priv->configured_uri, 
-				      account->source->url)) {
-			exchange_account_forget_password (priv->exchange_account);
-		}
-		else {
-			/* FIXME: modify esources and don't ask for re-login */
-			/* modify_esource (priv->exchange_account, 
-			 * account->source->url); 
-			 * return;
-			 */
-		}
+	if (requires_relogin (config_listener->priv->configured_uri, 
+			      account->source->url)) {
+		exchange_account_forget_password (priv->exchange_account);
+	} else if (strcmp (config_listener->priv->configured_name, account->name)) {
+		g_free (config_listener->priv->configured_name);
+		config_listener->priv->configured_name = g_strdup (account->name);
+		return;
+	} else {
+		/* FIXME: Do ESources need to be modified? */
+		return;
 	}
-	else if (strcmp (config_listener->priv->configured_name, account->name))
-		remove_account_esources (priv->exchange_account);
-
-	/* Nope. Let the user know we're ignoring him. */
-	e_notice (NULL, GTK_MESSAGE_WARNING,
-		  _("Changes to Exchange account configuration will "
-		    "take\nplace after you quit and restart Evolution."));
 
 	/* But note the new URI so if he changes something else, we
 	 * only warn again if he changes again.
 	 */
 	g_free (config_listener->priv->configured_uri);
 	config_listener->priv->configured_uri = g_strdup (account->source->url);
-	g_free (config_listener->priv->configured_name);
-	config_listener->priv->configured_name = g_strdup (account->name);
 }
 
 static void
@@ -561,9 +571,6 @@ account_removed (EAccountList *account_list, EAccount *account)
 	if (account != priv->configured_account)
 		return;
 
-	/* Remove all ESources */
-	remove_account_esources (priv->exchange_account);
-
 	exchange_account_forget_password (priv->exchange_account);
 
 	if (!exchange_account_get_context (priv->exchange_account)) {
@@ -582,17 +589,7 @@ account_removed (EAccountList *account_list, EAccount *account)
 		priv->configured_uri = NULL;
 		g_free (priv->configured_name);
 		priv->configured_name = NULL;
-	} else {
-		if (account->enabled) {
-			e_notice (NULL, GTK_MESSAGE_INFO,
-				_("The Exchange account will be removed when you quit Evolution"));
-		}
-		else {
-			/* The account is in use. We can't remove it. */
-			e_notice (NULL, GTK_MESSAGE_INFO,
-			  	_("The Exchange account will be disabled when you quit Evolution"));
-		}
-	}
+	} 
 }
 
 static gboolean
@@ -646,3 +643,52 @@ exchange_config_listener_get_accounts (ExchangeConfigListener *config_listener)
 	else
 		return NULL;
 }
+
+/**
+ * exchange_camel_urls_is_equal 
+ * 
+ * @url1: CAMEL URL string 1
+ * @url2: CAMEL URL string 2
+ *
+ * This function checks if the parameters present in two given CAMEL URLS are
+ * identical and returns the result.
+ *
+ * Return Value: Boolean result of the comparision.
+ *
+ **/ 
+static gboolean
+exchange_camel_urls_is_equal (const gchar *url1, const gchar *url2)
+{
+	CamelURL *curl1, *curl2;
+	gchar *param1, *param2;
+	const char *params[] = {
+		"auth",
+		"owa_url",
+		"owa_path",
+		"mailbox",
+		"ad_server",
+	};
+	const int n_params = 5;
+	int i;
+	
+	curl1 = camel_url_new (url1, NULL);
+	curl2 = camel_url_new (url2, NULL);
+
+	for (i = 0; i < n_params; ++i) {
+		param1 = (gchar*) camel_url_get_param (curl1, params[i]);
+		param2 = (gchar*) camel_url_get_param (curl2, params[i]);
+		if ((param1 && !param2) || (!param1 && param2) || /* Missing */
+		    (param1 && param2 && strcmp (param1, param2))) { /* Differing */
+			g_free (param1);
+			g_free (param2);
+			g_free (curl1);
+			g_free (curl2);
+			return FALSE;
+		}		
+		g_free (param1);
+		g_free (param2);
+	}
+	g_free (curl1);
+	g_free (curl2);
+	return TRUE;
+}
